package de.kif.common

import de.kif.common.model.ConstraintType
import de.kif.common.model.Schedule
import kotlinx.serialization.Serializable

@Serializable
data class ConstraintError(
    val reason: String = ""
)

@Serializable
data class ConstraintMap(
    val map: List<Pair<Long, List<ConstraintError>>>
)

fun checkConstraints(
    check: List<Schedule>,
    against: List<Schedule>
): ConstraintMap {
    val map = mutableMapOf<Long, List<ConstraintError>>()

    for (schedule in check) {
        if (schedule.id == null) continue
        val errors = mutableListOf<ConstraintError>()

        if (schedule.workGroup.projector && !schedule.room.projector) {
            errors += ConstraintError("Work group requires projector, but room does not have one!")
        }
        if (schedule.workGroup.internet && !schedule.room.internet) {
            errors += ConstraintError("Work group requires internet, but room does not have one!")
        }
        if (schedule.workGroup.whiteboard && !schedule.room.whiteboard) {
            errors += ConstraintError("Work group requires whiteboard, but room does not have one!")
        }
        if (schedule.workGroup.blackboard && !schedule.room.blackboard) {
            errors += ConstraintError("Work group requires blackboard, but room does not have one!")
        }
        if (schedule.workGroup.accessible && !schedule.room.accessible) {
            errors += ConstraintError("Work group requires accessible, but room does not have one!")
        }

        val start = schedule.getAbsoluteStartTime()
        val end = schedule.getAbsoluteEndTime()

        for (leader in schedule.workGroup.leader) {
            for (s in against) {
                if (
                    schedule != s &&
                    leader in s.workGroup.leader &&
                    start <= s.getAbsoluteEndTime() &&
                    s.getAbsoluteStartTime() <= end
                ) {
                    errors += ConstraintError("Work group with leader $leader cannot be at same time with ${s.workGroup.name}!")
                }
            }
        }

        for ((type, constraints) in schedule.workGroup.constraints.groupBy { it.type }) {
            when (type) {
                ConstraintType.OnlyOnDay -> {
                    val onlyOnDay = constraints.map { it.day == schedule.day }
                    if (onlyOnDay.none()) {
                        val dayList = constraints.mapNotNull { it.day }.distinct().sorted()
                        errors += ConstraintError("Work group requires days $dayList, but is on ${schedule.day}!")
                    }
                }

                ConstraintType.NotOnDay -> {
                    val notOnDay = constraints.map { it.day != schedule.day }
                    if (notOnDay.none()) {
                        val dayList = constraints.mapNotNull { it.day }.distinct().sorted()
                        errors += ConstraintError("Work group requires not days $dayList, but is on ${schedule.day}!")
                    }
                }

                ConstraintType.OnlyBeforeTime -> {
                    for (it in constraints) {
                        if (it.time == null) continue
                        if (it.day == null) {
                            if (it.time < schedule.time) {
                                errors += ConstraintError("Work group requires before time ${it.time}, but is on ${schedule.time}!")
                            }
                        } else {
                            if (it.day == schedule.day && it.time < schedule.time) {
                                errors += ConstraintError("Work group requires before time ${it.time} on day ${it.day}, but is on ${schedule.time}!")
                            }
                        }
                    }
                }

                ConstraintType.OnlyAfterTime -> {
                    for (it in constraints) {
                        if (it.time == null) continue
                        if (it.day == null) {
                            if (it.time > schedule.time) {
                                errors += ConstraintError("Work group requires after time ${it.time}, but is on ${schedule.time}!")
                            }
                        } else {
                            if (it.day == schedule.day && it.time > schedule.time) {
                                errors += ConstraintError("Work group requires after time ${it.time} on day ${it.day}, but is on ${schedule.time}!")
                            }
                        }
                    }
                }

                ConstraintType.NotAtSameTime -> {
                    for (constraint in constraints) {
                        for (s in against) {
                            if (
                                s.workGroup.id == constraint.workGroup &&
                                start <= s.getAbsoluteEndTime() &&
                                s.getAbsoluteStartTime() <= end
                            ) {
                                errors += ConstraintError("Work group requires not same time with ${s.workGroup.name}!")
                            }
                        }
                    }
                }

                ConstraintType.OnlyAfterWorkGroup -> {
                    for (constraint in constraints) {
                        for (s in against) {
                            if (
                                s.workGroup.id == constraint.workGroup &&
                                s.getAbsoluteEndTime() > start
                            ) {
                                errors += ConstraintError("Work group requires after ${s.workGroup.name}!")
                            }
                        }
                    }
                }
            }
        }

        map[schedule.id] = errors
    }

    return ConstraintMap(map.map { it.toPair() })
}
